/*
 *  collision2.mm
 *  iminib3d
 *
 *  Created by Simon Harrison.
 *  Copyright Si Design. All rights reserved.
 *
 */

#include "collision2.h"

#include "entity.h"
#include "pivot.h"
#include "mesh.h"
#include "camera.h"
#include "terrain.h"
#include "surface.h"
#include "collision.h"
#include "tree.h"
#include "geom.h"

#include <vector>
#include <list>
#include <iostream>

using namespace std;

extern Vector col_coords;

list<CollisionPair*> CollisionPair::cp_list;
list<Entity*> CollisionPair::ent_lists[MAX_TYPES];

 // dynamic to static
void UpdateCollisions(){
	UpdateStaticCollisions();
	UpdateDynamicCollisions();
}

int PositionEntities(int update_old,int add_to_new){
	int new_pos=false;
	list<CollisionPair*>::iterator cp_it;
	for(cp_it=CollisionPair::cp_list.begin();cp_it!=CollisionPair::cp_list.end();cp_it++){
		CollisionPair col_pair=**cp_it;

		// loop through src entities
		list<Entity*>::iterator src_ent_it;
		for(src_ent_it=CollisionPair::ent_lists[col_pair.src_type].begin();src_ent_it!=CollisionPair::ent_lists[col_pair.src_type].end();src_ent_it++){
			Entity& ent=**src_ent_it;
			if(ent.no_collisions!=0){
				if(update_old){
					ent.old_x=ent.EntityX(true);
					ent.old_y=ent.EntityY(true);
					ent.old_z=ent.EntityZ(true);
				}

				if(add_to_new){
					ent.new_x=(ent.new_x-ent.EntityX(true))+ent.EntityX(true);
					ent.new_y=(ent.new_y-ent.EntityY(true))+ent.EntityY(true);
					ent.new_z=(ent.new_z-ent.EntityZ(true))+ent.EntityZ(true);
				}
				ent.PositionEntity(ent.new_x,ent.new_y,ent.new_z);
				new_pos=true;
			}
		}
	}

	return new_pos;

}

void clearCollisions(){

	list<CollisionPair*>::iterator cp_it;

	for(cp_it=CollisionPair::cp_list.begin();cp_it!=CollisionPair::cp_list.end();cp_it++){

		CollisionPair col_pair=**cp_it;

		// loop through src entities
		list<Entity*>::iterator src_ent_it;

		for(src_ent_it=CollisionPair::ent_lists[col_pair.src_type].begin();src_ent_it!=CollisionPair::ent_lists[col_pair.src_type].end();src_ent_it++){

			Entity& ent=**src_ent_it;

			ent.no_collisions=0;
			for(unsigned int ix=0;ix<ent.collision.size();ix++){
				delete ent.collision[ix];
			}
			ent.collision.clear();
		}
	}
}

void UpdateStaticCollisions(){
	list<CollisionPair*>::iterator cp_it;
	for(cp_it=CollisionPair::cp_list.begin();cp_it!=CollisionPair::cp_list.end();cp_it++){
		CollisionPair col_pair=**cp_it;

		// if no entities exist of src_type or des_type then do not check for collisions
		if((CollisionPair::ent_lists[col_pair.src_type].size()==0) ||
       (CollisionPair::ent_lists[col_pair.des_type].size()==0))
      continue;

		// loop through src entities
		list<Entity*>::iterator src_ent_it;

		for(src_ent_it=CollisionPair::ent_lists[col_pair.src_type].begin();src_ent_it!=CollisionPair::ent_lists[col_pair.src_type].end();src_ent_it++){
			Entity& ent1 = **src_ent_it;
			// clear collisions
			ent1.no_collisions=0;
			for(unsigned int ix=0;ix<ent1.collision.size();ix++){
				delete ent1.collision[ix];
			}
			ent1.collision.clear();

			// if src entity is hidden or it's parent is hidden then do not check for collision
			if(ent1.Hidden()==true)
			  continue;

			Vector c_vec_a(ent1.EntityX(true),
                     ent1.EntityY(true),
                     ent1.EntityZ(true));
			Vector c_vec_b(ent1.old_x,
                     ent1.old_y,
                     ent1.old_z);
			Vector c_vec_radius(ent1.radius_x,
                          ent1.radius_y,
                          ent1.radius_x);

			CollisionInfo* c_col_info = C_CreateCollisionInfoObject(&c_vec_a,&c_vec_b,&c_vec_radius);

			Collision* c_coll=NULL;

			int response=0;

			// repeat until there's no collision between src and dest entities
			for(;;){
				int hit=false;

				c_coll = C_CreateCollisionObject();
				Entity* ent2_hit = NULL;

				// loop through des entities that are paired with src entities
				list<Entity*>::iterator des_ent_it;
				for(des_ent_it=CollisionPair::ent_lists[col_pair.des_type].begin();des_ent_it!=CollisionPair::ent_lists[col_pair.des_type].end();des_ent_it++) {
					Entity& ent2=**des_ent_it;

					// if src ent is same as des entity then do not check for collision
					if (&ent1==&ent2)
					  continue;

					// if des entity is hidden or it's parent is hidden then do not check for collision
					if (ent2.Hidden()==true)
					  continue;

					//if(QuickCheck(ent,ent2)==false) continue; // quick check to see if entities are colliding
					Matrix mat;

					if(ent2.dynamic != true){
						mat.Overwrite(ent2.mat);
					}else{
						mat.Overwrite(ent2.old_mat);
					}

					Vector c_vec_i( mat.grid[0][0], mat.grid[0][1],-mat.grid[0][2]);
					Vector c_vec_j( mat.grid[1][0], mat.grid[1][1],-mat.grid[1][2]);
					Vector c_vec_k(-mat.grid[2][0],-mat.grid[2][1], mat.grid[2][2]);

					MMatrix c_mat(c_vec_i,c_vec_j,c_vec_k);
					Vector c_vec_v(mat.grid[3][0],mat.grid[3][1],-mat.grid[3][2]);

					Transform c_tform(c_mat,c_vec_v);

					// if pick mode is sphere or box then update collision info object
					// to include entity radius/box info
					if(col_pair.col_method != COLLISION_METHOD_POLYGON){
						C_UpdateCollisionInfoObject(c_col_info,
                                        ent2.radius_x,
                                        ent2.box_x,
                                        ent2.box_y,
                                        ent2.box_z,
                                        ent2.box_x+ent2.box_w,
                                        ent2.box_y+ent2.box_h,
                                        ent2.box_z+ent2.box_d);
					}

					MeshCollider* tree = NULL;
					if(dynamic_cast<Mesh*>(&ent2)!=NULL) {
						Mesh* m=dynamic_cast<Mesh*>(&ent2);
						// create collision tree for mesh if necessary
						m->TreeCheck();
						tree = m->c_col_tree;
					} else if(dynamic_cast<Terrain*>(&ent2)!=NULL) {
						Terrain* t=dynamic_cast<Terrain*>(&ent2);
						// create collision tree for terrain if necessary
						t->TreeCheck(c_col_info);
						tree = t->c_col_tree;
					}

					hit = C_CollisionDetect(c_col_info,c_coll,&c_tform,tree,col_pair.col_method);
					if (hit) {
						ent2_hit=&ent2;
					}
					response = col_pair.response;

				} // end of dest ent loop

				if(ent2_hit){

					ent1.no_collisions += 1;

					//int i=ent.no_collisions-1;
					CollisionImpact* eci=new CollisionImpact;
					ent1.collision.push_back(eci);

					eci->x   = C_CollisionX();
					eci->y   = C_CollisionY();
					eci->z   = C_CollisionZ();

					eci->nx  = C_CollisionNX();
					eci->ny  = C_CollisionNY();
					eci->nz  = C_CollisionNZ();

					eci->ent = ent2_hit;

					if(dynamic_cast<Mesh*>(ent2_hit)!=NULL){
						eci->surf=dynamic_cast<Mesh*>(ent2_hit)->GetSurface(C_CollisionSurface());
					} else {
						eci->surf=NULL;
					}

					eci->tri = C_CollisionTriangle();

					if (C_CollisionResponse(c_col_info,c_coll,response)==false)
					  break;

				} else {
					break;
				}

				C_DeleteCollisionObject(c_coll);

			} // end of infinite loop

			C_DeleteCollisionObject(c_coll);

			int hits = C_CollisionFinal(c_col_info);

			if (hits) {

				ent1.new_x = C_CollisionPosX();
				ent1.new_y = C_CollisionPosY();
				ent1.new_z = C_CollisionPosZ();

				// moved from PositionEntities
				ent1.PositionEntity(ent1.new_x,ent1.new_y,ent1.new_z,true);
			}
			C_DeleteCollisionInfoObject(c_col_info);
		} // end of src ent loop
	} // end of collision pair loop
}

// dynamic to dynamic
void UpdateDynamicCollisions(){

	Vector c_vec_i(1.0f, 0.0f, 0.0f);
	Vector c_vec_j(0.0f, 1.0f, 0.0f);
	Vector c_vec_k(0.0f, 0.0f, 1.0f);

	MMatrix c_mat(c_vec_i, c_vec_j, c_vec_k);
	Vector c_vec_v(0.0, 0.0, 0.0);

	Transform c_tform(c_mat, c_vec_v);

	static Pivot* piv1o  = Pivot::CreatePivot();
	static Pivot* piv1   = Pivot::CreatePivot(piv1o);
	static Pivot* piv11  = Pivot::CreatePivot(piv1o);
	static Pivot* piv111 = Pivot::CreatePivot(piv1o);

	static Pivot* piv2o = Pivot::CreatePivot();
	static Pivot* piv2  = Pivot::CreatePivot(piv2o);

	float xx,xy,xz;
	float yx,yy,yz;
	float zx,zy,zz;
	list<CollisionPair*>::iterator cp_it;

	for(cp_it=CollisionPair::cp_list.begin(); cp_it!=CollisionPair::cp_list.end(); cp_it++) {
		CollisionPair col_pair=**cp_it;

		// if no entities exist of src_type or des_type then do not check for collisions
		if ( (CollisionPair::ent_lists[col_pair.src_type].size()==0) ||
         (CollisionPair::ent_lists[col_pair.des_type].size()==0) )
      continue;

		// loop through src entities
		list<Entity*>::iterator src_ent_it;
		for ( src_ent_it=CollisionPair::ent_lists[col_pair.src_type].begin(); src_ent_it!=CollisionPair::ent_lists[col_pair.src_type].end(); src_ent_it++) {
    	Entity& ent1=**src_ent_it;

			// if src entity is hidden or it's parent is hidden then do not check for collision
			if(ent1.Hidden()==true)
			  continue;

			// loop through des entities that are paired with src entities
			list<Entity*>::iterator des_ent_it;
			for( des_ent_it=CollisionPair::ent_lists[col_pair.des_type].begin(); des_ent_it!=CollisionPair::ent_lists[col_pair.des_type].end(); des_ent_it++) {
				Entity& ent2=**des_ent_it;

				if (ent2.dynamic!=true)
				  continue;

				// if des entity is hidden or it's parent is hidden then do not check for collision
				if (ent2.Hidden()==true)
				  continue;

				// if src ent is same as des entity then do not check for collision
				if (&ent1==&ent2)
				  continue;

        /* !!! changed
				float dx1 = ent1.EntityX(true) - ent2.EntityX(true);
				float dy1 = ent1.EntityY(true) - ent2.EntityY(true);
				float dz1 = ent1.EntityZ(true) - ent2.EntityZ(true);
        piv1->PositionEntity(dx1, dy1 ,dz1, false);

				float dx2 = ent1.old_x - ent2.old_x;
				float dy2 = ent1.old_y - ent2.old_y;
				float dz2 = ent1.old_z - ent2.old_z;
				piv2->PositionEntity(dx2, dy2 ,dz2, false);
        */


        piv1->PositionEntity(ent1.EntityX(true) - ent2.EntityX(true),
                             ent1.EntityY(true) - ent2.EntityY(true),
                             ent1.EntityZ(true) - ent2.EntityZ(true), false);

				piv2->PositionEntity(ent1.old_x - ent2.old_x,
                             ent1.old_y - ent2.old_y,
                             ent1.old_z - ent2.old_z, false);

				ent2.mat.GetInverse2(piv1o->mat);
				piv1o->mat.SetTranslate(0,0,0);
				piv1->MQ_Update();

				piv11->MQ_Update();
				piv111->MQ_Update();

				ent2.old_mat.GetInverse2(piv2o->mat);
				piv2o->mat.SetTranslate(0,0,0);
				piv2->MQ_Update();

        // !!! changed
				xx=1.0f; xy=0.0f; xz=0.0f;
				yx=0.0f; yy=1.0f; yz=0.0f;
				zx=0.0f; zy=0.0f; zz=1.0f;

				ent2.mat.TransformVec(xx, xy, xz);
				ent2.mat.TransformVec(yx, yy, yz);
				ent2.mat.TransformVec(zx, zy, zz);
        // !!! changed to float math
				piv1o->sx = 1.0f/sqrtf(xx*xx + xy*xy + xz*xz );
				piv1o->sy = 1.0f/sqrtf(yx*yx + yy*yy + yz*yz );
				piv1o->sz = 1.0f/sqrtf(zx*zx + zy*zy + zz*zz );

				Vector vec_a(piv1->EntityX(true), piv1->EntityY(true), piv1->EntityZ(true));
				Vector vec_b(piv2->EntityX(true), piv2->EntityY(true), piv2->EntityZ(true));
				Vector vec_radius(ent1.radius_x*piv1o->sx,
                          ent1.radius_x*piv1o->sy,
                          ent1.radius_x*piv1o->sz);

				CollisionInfo* c_col_info = C_CreateCollisionInfoObject(&vec_a,&vec_b,&vec_radius);

				MeshCollider* tree = NULL;
				if(dynamic_cast<Mesh*>(&ent2)!=NULL){
					Mesh* m=dynamic_cast<Mesh*>(&ent2);
					// create collision tree for mesh if necessary
					m->TreeCheck();
					tree=m->c_col_tree;
				}

				Collision* c_coll=NULL;
				// repeat until there's no collision between src and dest entities
				for(;;){
					c_coll = C_CreateCollisionObject();
					// method set to 2
					int hit = C_CollisionDetect(c_col_info,c_coll,&c_tform,tree,2);
					if (hit) {
						if (C_CollisionResponse(c_col_info,c_coll,col_pair.response)==false)
						  break;
					} else {
						break;
					}
					C_DeleteCollisionObject(c_coll);
				}

				C_DeleteCollisionObject(c_coll);

				int hits = C_CollisionFinal(c_col_info);

				if (hits) {
					// register collision
					ent1.no_collisions += 1;

					// int i=ent.no_collisions-1;
					CollisionImpact* eci = new CollisionImpact;
					ent1.collision.push_back(eci);

					// !!! will be overwritten by piv11,piv2
					eci->x   = C_CollisionX();
					eci->y   = C_CollisionY();
					eci->z   = C_CollisionZ();
					eci->nx  = C_CollisionNX();
					eci->ny  = C_CollisionNY();
					eci->nz  = C_CollisionNZ();

					eci->ent = &ent2;

					if(dynamic_cast<Mesh*>(&ent2)!=NULL){
						eci->surf = dynamic_cast<Mesh*>(&ent2)->GetSurface(C_CollisionSurface());
					}else{
						eci->surf = NULL;
					}

					eci->tri = C_CollisionTriangle();

					piv1o->RotateEntity(0.0f,0.0f,0.0f);

					/* !!! changed
					float x = C_CollisionPosX();
					float y = C_CollisionPosY();
					float z = C_CollisionPosZ();*/

					piv1->PositionEntity(C_CollisionPosX(),C_CollisionPosY(),C_CollisionPosZ(),true);
					piv11->PositionEntity(eci->x,eci->y,eci->z,true);

					piv2o->mat.LoadIdentity();
					piv2->PositionEntity(eci->nx,eci->ny,eci->nz,false);

					piv1o->mat.Overwrite(ent2.mat);
					piv1o->mat.Scale(piv1o->sx,piv1o->sy,piv1o->sz);
					piv1->MQ_Update();
					piv11->MQ_Update();

					piv2o->mat.Overwrite(ent2.mat);
					piv2o->mat.SetTranslate(0.0f,0.0f,0.0f);
					piv2->MQ_Update();

					// update stored collision impact values
					eci->x  = piv11->EntityX(true);
					eci->y  = piv11->EntityY(true);
					eci->z  = piv11->EntityZ(true);

					eci->nx = piv2->EntityX(true);
					eci->ny = piv2->EntityY(true);
					eci->nz = piv2->EntityZ(true);

					/* !!! changed
					x = piv1->EntityX(true);
					y = piv1->EntityY(true);
					z = piv1->EntityZ(true);*/
					ent1.new_x = piv1->EntityX(true);
					ent1.new_y = piv1->EntityY(true);
					ent1.new_z = piv1->EntityZ(true);
					ent1.PositionEntity(ent1.new_x, ent1.new_y, ent1.new_z,true);
				}

				// reset
  			C_DeleteCollisionInfoObject(c_col_info);

				piv2->PositionEntity(0.0f,0.0f,0.0f,true);
				piv2->RotateEntity  (0.0f,0.0f,0.0f,true);

				piv1->PositionEntity(0.0f,0.0f,0.0f,true);
				piv1->RotateEntity  (0.0f,0.0f,0.0f,true);

				piv11->PositionEntity(0.0f,0.0f,0.0f,true);
				piv11->RotateEntity  (0.0f,0.0f,0.0f,true);

				piv111->PositionEntity(0.0f,0.0f,0.0f,true);
				piv111->RotateEntity  (0.0f,0.0f,0.0f,true);

				piv1o->PositionEntity(0.0f,0.0f,0.0f,true);
				piv1o->RotateEntity  (0.0f,0.0f,0.0f,true);
				piv1o->ScaleEntity   (1.0f,1.0f,1.0f,true);

				piv2o->PositionEntity(0.0f,0.0f,0.0f,true);
				piv2o->RotateEntity  (0.0f,0.0f,0.0f,true);
				piv2o->ScaleEntity   (1.0f,1.0f,1.0f,true);

			} // end of dest ent loop

		} // end of src ent loop

	} // end of collision pairs loop

}

// perform quick check to see whether it is possible that ent and ent 2 are intersecting
/* !!! not used
int QuickCheck(Entity& ent,Entity& ent2){
	// check to see if src ent has moved since last update - if not, no intersection
	if (ent.old_x==ent.EntityX(true) &&
      ent.old_y==ent.EntityY(true) &&
      ent.old_z==ent.EntityZ(true)) {
		return false;
	}
	return true;
}*/
