#ifdef EMSCRIPTEN
#include <GLES2/gl2.h>
#define GLES2
#endif

#include "global.h"

#include "texture.h"
#include "entity.h"
#include "mesh.h"

#include "camera.h"
#include "pick.h"
#include "light.h"
#include "shadow.h"
#include "stencil.h"
#include "csg.h"
#include "voxel.h"
#include "octree.h"
#include "geosphere.h"
#include "isosurface.h"
#include "particle.h"
#include "physics.h"
#include "actions.h"
#include "postfx.h"

#ifndef GL_FRAGMENT_SHADER
#define GL_FRAGMENT_SHADER 0x8B30
#endif

#ifndef GL_VERTEX_SHADER
#define GL_VERTEX_SHADER 0x8B31
#endif

#ifndef GL_GEOMETRY_SHADER
#define GL_GEOMETRY_SHADER 0x8DD9
#endif

#ifndef GL_MULTISAMPLE
#define GL_MULTISAMPLE 0x809D
#endif

#include "functions.h"

#ifdef __cplusplus
extern "C" {
#endif

void BufferToTex(Texture* tex,unsigned char* buffer, int frame){
	tex->BufferToTex(buffer,frame);
}

void BackBufferToTex(Texture* tex,int frame){
	tex->BackBufferToTex(frame);
}

void CameraToTex(Texture* tex, Camera* cam, int frame){
	tex->CameraToTex(cam,frame);
}

void TexToBuffer(Texture* tex,unsigned char* buffer, int frame){
	tex->TexToBuffer(buffer,frame);
}


/*
bbdoc: Minib3d Only
about:

This command is the equivalent of Blitz3D's MeshCullBox command.

It is used to set the radius of a mesh's 'cull sphere'
if the 'cull sphere' is not inside the viewing area, the mesh will not be rendered.

A mesh's cull radius is set automatically, therefore
in most cases you will not have to use this command.

One time you may have to use it is for animated meshes
where the default cull radius may not take into account all animation
positions, resulting in the mesh being wrongly culled at extreme positions.
*/
void MeshCullRadius(Entity* ent, float radius) {
  ent->cull_radius=-radius;
}

int AddAnimSeq(Entity* ent,int length){
	return ent->AddAnimSeq(length);
}

void AddMesh(Mesh* mesh1,Mesh* mesh2){
  mesh1->AddMesh(mesh2);
}

int AddTriangle(Surface* surf,int v0,int v1,int v2){
  return surf->AddTriangle(v0,v1,v2);
}


int AddVertex(Surface* surf,float x, float y,float z,float u, float v, float w){
  return surf->AddVertex(x,y,z,u,v,w);
}

void AmbientLight(float r,float g,float b){
  Global::AmbientLight(r,g,b);
}

void AntiAlias(int bOnOff) {
#ifndef GLES2
	if (bOnOff!=0){
		glEnable(GL_MULTISAMPLE);
	}else{
	  glDisable(GL_MULTISAMPLE);
	}
#endif
}

void Animate(Entity* ent,int mode=1,float speed=1.0f,int seq=0,int trans=0){
	ent->Animate(mode,speed,seq,trans);
}

int Animating(Entity* ent){
	if (ent->anim_mode!=0) // !=ANIM_STOP
	  return true;

	if (ent->anim_trans!=0)
	  return true;

	return false;
}

int AnimLength(Entity* ent){
	return ent->AnimLength();
}

int AnimSeq(Entity* ent){
  return ent->anim_seq;
}

float AnimTime(Entity* ent) {
  return ent->AnimTime();
}

void BrushAlpha(Brush* brush, float a) {
  brush->BrushAlpha(a);
}

void BrushBlend(Brush* brush,int blend) {
  brush->BrushBlend(blend);
}

void BrushColor(Brush* brush,float r,float g,float b) {
  brush->BrushColor(r,g,b);
}

void BrushFX(Brush* brush,int fx) {
  brush->BrushFX(fx);
}

void BrushShininess(Brush* brush,float s) {
  brush->BrushShininess(s);
}

void BrushTexture(Brush* brush,Texture* tex,int frame,int index) {
  brush->BrushTexture(tex,frame,index);
}

void CameraClsColor(Camera* cam, float r,float g,float b){
	cam->CameraClsColor(r,g,b);
}

void CameraClsMode(Camera* cam,int bClearColorBuffer,int bClearDepthBuffer){
	cam->CameraClsMode( (bClearColorBuffer!=0),(bClearDepthBuffer!=0) );
}

void CameraFogColor(Camera* cam,float r,float g,float b){
	cam->CameraFogColor(r,g,b);
}

void CameraFogMode(Camera* cam,int bMode){
	cam->CameraFogMode((bMode!=0));
}

void CameraFogRange(Camera* cam,float nnear,float nfar){
	cam->CameraFogRange(nnear,nfar);
}

Entity* CameraPick(Camera* cam,float x,float y){
	return Pick::CameraPick(cam,x,y);
}

void CameraProject(Camera* cam,float x,float y,float z){
	cam->CameraProject(x,y,z);
}

void  CameraProjMode(Camera* cam,int mode){
	cam->CameraProjMode(mode);
}

void  CameraRange(Camera* cam,float nnear,float nfar){
	cam->CameraRange(nnear,nfar);
}

void  CameraViewport(Camera* cam,int x,int y,int width,int height){
	cam->CameraViewport(x,y,width,height);
}

void  CameraZoom(Camera* cam,float zoom){
	cam->CameraZoom(zoom);
}

void ClearCollisions(){
	Global::ClearCollisions();
}

void ClearSurface(Surface* surf,int clear_verts,int clear_tris){
	surf->ClearSurface(clear_verts,clear_tris);
}

void ClearTextureFilters(){
	Texture::ClearTextureFilters();
}

void ClearWorld(int bEntities,int bTextures){
	Global::ClearWorld(bEntities,0,bTextures);
}

Entity* CollisionEntity(Entity* ent,int index){
	return ent->CollisionEntity(index);
}

void Collisions(int src_no,int dest_no,int method_no,int response_no){
	Global::Collisions(src_no,dest_no,method_no,response_no);
}

float CollisionNX(Entity* ent,int index){
	return ent->CollisionNX(index);
}

float CollisionNY(Entity* ent,int index){
	return ent->CollisionNY(index);
}

float CollisionNZ(Entity* ent,int index){
	return ent->CollisionNZ(index);
}

Surface* CollisionSurface(Entity* ent,int index){
	return ent->CollisionSurface(index);
}

float CollisionTime(Entity* ent,int index){
	return ent->CollisionTime(index);
}

int CollisionTriangle(Entity* ent,int index){
	return ent->CollisionTriangle(index);
}

float CollisionX(Entity* ent,int index){
	return ent->CollisionX(index);
}

float CollisionY(Entity* ent,int index){
	return ent->CollisionY(index);
}

float CollisionZ(Entity* ent,int index){
	return ent->CollisionZ(index);
}

int CountChildren(Entity* ent){
	return ent->CountChildren();
}

int CountCollisions(Entity* ent){
	return ent->CountCollisions();
}

Entity* CopyEntity(Entity* ent,Entity* parent){
	return ent->CopyEntity(parent);
}

Mesh* CopyMesh(Mesh* mesh,Entity* parent){
	return mesh->CopyMesh(parent);
}

int CountSurfaces(Mesh* mesh){
	return mesh->CountSurfaces();
}

int CountTriangles(Surface* surf){
	return surf->CountTriangles();
}

int CountVertices(Surface* surf){
	return surf->CountVertices();
}

Blob* CreateBlob(Fluid* fluid, float radius, Entity* parent_ent){
	return Blob::CreateBlob(fluid, radius, parent_ent);
}

Bone* CreateBone(Mesh* mesh, Entity* parent_ent){
	return mesh->CreateBone(parent_ent);
}

Brush* CreateBrush(float r,float g,float b){
	return Brush::CreateBrush(r,g,b);
}

Camera* CreateCamera(Entity* parent){
	return Camera::CreateCamera(parent);
}

Constraint* CreateConstraint(Entity* p1, Entity* p2, float l){
	return Constraint::CreateConstraint(p1, p2, l);
}

Mesh* CreateCone(int segments, int solid, Entity* parent){
	return Mesh::CreateCone(segments,(solid!=0),parent);
}

Mesh* CreateCylinder(int segments,int solid,Entity* parent){
	return Mesh::CreateCylinder(segments,(solid!=0),parent);
}

Mesh* CreateCube(Entity* parent){
	return Mesh::CreateCube(parent);
}

Terrain* CreateGeosphere(int size, Entity* parent){
	return Geosphere::CreateGeosphere(size,parent);
}

Mesh* CreateMesh(Entity* parent){
	return Mesh::CreateMesh(parent);
}

Light* CreateLight(int light_type,Entity* parent){
	return Light::CreateLight(light_type,parent);
}

Pivot* CreatePivot(Entity* parent){
	return Pivot::CreatePivot(parent);
}

Mesh* CreatePlane(Entity* parent){
	return Mesh::CreatePlane(parent);
}

Mesh* CreateQuad(Entity* parent){
    return Mesh::CreateQuad(parent);
}

RigidBody* CreateRigidBody(Entity* body, Entity* p1, Entity* p2, Entity* p3, Entity* p4){
	return RigidBody::CreateRigidBody(body, p1, p2, p3, p4);
}

ShadowObject* CreateShadow(Mesh* parent, int bStatic, Mesh* mesh2){
	return ShadowObject::Create(parent, (bStatic!=0), mesh2);
}

Mesh* CreateSphere(int segments,Entity* parent){
	return Mesh::CreateSphere(segments,parent);
}

Sprite* CreateSprite(Entity* parent){
	return Sprite::CreateSprite(parent);
}

Surface* CreateSurface(Mesh* mesh,Brush* brush){
	return mesh->CreateSurface(brush);
}

Stencil* CreateStencil(){
	return Stencil::CreateStencil();
}

Terrain* CreateTerrain(int size, Entity* parent){
	return Terrain::CreateTerrain(size,parent);
}

Texture* CreateTexture(int width,int height,int flags){
	return Texture::CreateTexture(width,height,flags);
}

VoxelSprite* CreateVoxelSprite(int slices, Entity* parent){
	return VoxelSprite::CreateVoxelSprite(slices, parent);
}

float DeltaPitch(Entity* ent1,Entity* ent2){
	return ent1->DeltaPitch(ent2);
}

float DeltaYaw(Entity* ent1,Entity* ent2){
	return ent1->DeltaYaw(ent2);
}

void EmitterVector(ParticleEmitter* emit, float x, float y, float z){
	emit->EmitterVector(x, y, z);
}

void EmitterRate(ParticleEmitter* emit, float rate){
	emit->EmitterRate(rate);
}

void EmitterParticleLife(ParticleEmitter* emit, int l){
	emit->EmitterParticleLife(l);
}

void EmitterParticleFunction(ParticleEmitter* emit, void (*EmitterFunction)(Entity*, int)){
	emit->EmitterParticleFunction(EmitterFunction);
}

void EmitterParticleSpeed(ParticleEmitter* emit, float s){
	emit->EmitterParticleSpeed(s);
}

void EmitterVariance(ParticleEmitter* emit, float v){
	emit->EmitterVariance(v);
}

void EntityAlpha(Entity* ent,float alpha, int bRecursive){
	ent->EntityAlpha(alpha,(bRecursive!=0));
}

/*void EntityAutoFade(Entity* ent,float near,float far){}*/

void EntityBlend(Entity* ent, int blend,int bRecursive){
	ent->EntityBlend(blend, (bRecursive!=0));
}

void EntityBox(Entity* ent,float x,float y,float z,float w,float h,float d){
	ent->EntityBox(x,y,z,w,h,d);
}

const char* EntityClass(Entity* ent){
	return ent->EntityClass().c_str();
}

Entity* EntityCollided(Entity* ent,int type_no){
	return ent->EntityCollided(type_no);
}

void EntityColor(Entity* ent,float red,float green,float blue,int bRecursive){
	ent->EntityColor(red,green,blue,(bRecursive!=0));
}

float EntityDistance(Entity* ent1,Entity* ent2){
	return ent1->EntityDistance(ent2);
}

void EntityFX(Entity* ent, int fx, int bRecursive){
	ent->EntityFX(fx, (bRecursive!=0));
}

int EntityInView(Entity* ent,Camera* cam){
	return cam->EntityInView(ent);
}

const char* EntityName(Entity* ent){
	return ent->EntityName().c_str();
}

void EntityOrder(Entity* ent,int order,int bRrecursive){
	ent->EntityOrder(order, bRrecursive);
}

void EntityParent(Entity* ent,Entity* parent_ent,int glob){
	ent->EntityParent(parent_ent,(glob!=0));
}

Entity* EntityPick(Entity* ent,float range){
	return Pick::EntityPick(ent,range);
}

void EntityPickMode(Entity* ent,int pick_mode){
	ent->EntityPickMode(pick_mode);
}

float EntityPitch(Entity* ent,int glob){
	return ent->EntityPitch((glob!=0));
}

void EntityRadius(Entity* ent,float radius_x,float radius_y){
	ent->EntityRadius(radius_x,radius_y);
}

float EntityRoll(Entity* ent,int glob){
	return ent->EntityRoll( (glob!=0) );
}

void EntityShininess(Entity* ent,float shine, int bRecursive){
	ent->EntityShininess(shine,(bRecursive!=0));
}

void EntityTexture(Entity* ent,Texture* tex,int frame,int index, int bRecursive){
	ent->EntityTexture(tex,frame,index,(bRecursive!=0));
}

void EntityType(Entity* ent,int type_no,int bRecursive){
	ent->EntityType(type_no,(bRecursive!=0));
}

int EntityVisible(Entity* src_ent,Entity* dest_ent){
	return Pick::EntityVisible(src_ent,dest_ent);
}

float EntityX(Entity* ent, int glob){
	return ent->EntityX( (glob!=0) );
}

float EntityY(Entity* ent,int glob){
	return ent->EntityY( (glob!=0) );
}

float EntityYaw(Entity* ent,int glob){
	return ent->EntityYaw( (glob!=0) );
}

float EntityZ(Entity* ent,int glob){
	return ent->EntityZ( (glob!=0) );
}

int ExtractAnimSeq(Entity* ent,int first_frame,int last_frame,int seq){
	return ent->ExtractAnimSeq(first_frame,last_frame,seq);
}

Entity* FindChild(Entity* ent,char* child_name){
	return ent->FindChild(child_name);
}

void FitMesh(Mesh* mesh,float x,float y,float z,float width,float height,float depth,int uniform){
	mesh->FitMesh(x,y,z,width,height,depth,(uniform!=0));
}

void FlipMesh(Mesh* mesh){
	mesh->FlipMesh();
}

void FluidArray(Fluid* fluid, float* Array, int w, int h, int d){
	fluid->FluidArray(Array, w, h, d);
}

void FluidFunction(Fluid* fluid, float (*FieldFunction)(float, float, float)){
	fluid->FluidFunction(FieldFunction);
}

void FluidThreshold(Fluid* fluid, float threshold){
	fluid->threshold=threshold;
}

/*
bbdoc: <a href="http://www.blitzbasic.com/b3ddocs/command.php?name=FreeBrush">Online Help</a>
*/
void  FreeBrush(Brush* brush){
	brush->FreeBrush();
}

void FreeConstraint(Constraint* con){
	con->FreeConstraint();
}

void FreeEntity(Entity* ent){
	ent->FreeEntity();
}

void FreeRigidBody(RigidBody* body){
	body->FreeRigidBody();
}

void FreeShadow(ShadowObject* shad){
	shad->FreeShadow();
}

void FreeTexture(Texture* tex){
	tex->FreeTexture();
}

void GeosphereHeight(Geosphere* geo, float h){
	geo->vsize=h;
}

Texture* GetBrushTexture(Brush* brush,int index){
  return brush->GetBrushTexture(index);
}

Entity* GetChild(Entity* ent,int child_no){
	return ent->GetChild(child_no);
}

Brush* GetEntityBrush(Entity* ent){
	return ent->GetEntityBrush();
}

int GetEntityType(Entity* ent){
	return ent->GetEntityType();
}

Entity* GetParentEntity(Entity* ent){
	return ent->GetParent();
}

Surface* GetSurface(Mesh* mesh,int surf_no){
	return mesh->GetSurface(surf_no);
}

Brush* GetSurfaceBrush(Surface* surf){
	return surf->GetSurfaceBrush();
}

void Graphics3D(int width,int height){
	Global::Graphics(); //(width,height,depth,mode,rate);
	Global::width=width;
	Global::height=height;
}

void HandleSprite(Sprite* sprite,float h_x,float h_y){
	sprite->HandleSprite(h_x,h_y);
}

void HideEntity(Entity* ent){
	ent->HideEntity();
}

void LightColor(Light* light,float red,float green,float blue){
	light->LightColor(red,green,blue);
}

void LightConeAngles(Light* light,float inner_ang,float outer_ang){
	light->LightConeAngles(inner_ang,outer_ang);
}

void LightRange(Light* light,float range){
	light->LightRange(range);
}

Entity* LinePick(float x,float y,float z,float dx,float dy,float dz,float radius){
	return Pick::LinePick(x,y,z,dx,dy,dz,radius);
}

Mesh* LoadAnimMesh(char* file,Entity* parent){
	return Mesh::LoadAnimMesh(file,parent);
}

int LoadAnimSeq(Entity* ent, char* file){
	return ent->LoadAnimSeq(file);
}

Texture* LoadAnimTexture(char* file,int flags,int frame_width,int frame_height,int first_frame,int frame_count){
	return Texture::LoadAnimTexture(file,flags,frame_width,frame_height,first_frame,frame_count);
}

Brush* LoadBrush(char *file,int flags,float u_scale,float v_scale){
	return Brush::LoadBrush(file,flags,u_scale,v_scale);
}

Terrain* LoadGeosphere(char* file,Entity* parent){
	return Geosphere::LoadGeosphere(file,parent);
}

Mesh* LoadMesh(char* file,Entity* parent){
	return Mesh::LoadMesh(file,parent);
}

Terrain* LoadTerrain(char* file,Entity* parent){
	return Terrain::LoadTerrain(file,parent);
}

Texture* LoadTexture(char* file,int flags){
	return Texture::LoadTexture(file,flags);
}

Sprite* LoadSprite(char* tex_file,int tex_flag,Entity* parent){
	return Sprite::LoadSprite(tex_file,tex_flag,parent);
}

Mesh* MeshCSG(Mesh* m1, Mesh* m2, int method = 1){
	return CSG::MeshCSG(m1, m2, method);
}

float MeshDepth(Mesh* mesh){
	return mesh->MeshDepth();
}

int MeshesIntersect(Mesh* mesh1,Mesh* mesh2){
	return mesh1->MeshesIntersect(mesh2);
}

float MeshHeight(Mesh* mesh){
	return mesh->MeshHeight();
}

float MeshWidth(Mesh* mesh){
	return mesh->MeshWidth();
}

void MeshColor(Mesh* mesh,float r,float g,float b,float a) {
  if(mesh) mesh->MeshColor(r,g,b,a);
}
void MeshRed(Mesh* mesh,float r) {
  if(mesh) mesh->MeshRed(r);
}
void MeshGreen(Mesh* mesh,float g) {
  if(mesh) mesh->MeshGreen(g);
}
void MeshBlue(Mesh* mesh,float b) {
  if(mesh) mesh->MeshBlue(b);
}
void MeshAlpha(Mesh* mesh,float a) {
  if(mesh) mesh->MeshAlpha(a);
}


void ModifyGeosphere(Geosphere* geo, int x, int z, float new_height){
	geo->ModifyGeosphere ( x,  z, new_height);
}


void ModifyTerrain(Terrain* terr, int x, int z, float new_height){
	terr->ModifyTerrain ( x,  z, new_height);
}

void MoveEntity(Entity* ent,float x,float y,float z){
	ent->MoveEntity(x,y,z);
}

void NameEntity(Entity* ent,char* name){
	ent->NameEntity(name);
}

void PaintEntity(Entity* ent,Brush* brush){
	ent->PaintEntity(*brush);
}

void PaintMesh(Mesh* mesh,Brush* brush){
	mesh->PaintMesh(brush);
}

void PaintSurface(Surface* surf,Brush* brush){
	surf->PaintSurface(brush);
}

void ParticleColor(Sprite* sprite, float r, float g, float b, float a) {
  if (sprite!=NULL) {
	  ParticleBatch* p = ParticleBatch::GetParticleBatch(sprite->brush.tex[0],sprite->brush.blend,sprite->order);
	  if (p!=NULL) {
	    p->brush.red   = r+2;
	    p->brush.green = g+2;
	    p->brush.blue  = b+2;
	    p->brush.alpha = a+2;
	  }
	}
}

void ParticleVector(Sprite* sprite, float x, float y, float z) {
  if (sprite!=NULL) {
	  ParticleBatch* p=ParticleBatch::GetParticleBatch(sprite->brush.tex[0],sprite->brush.blend,sprite->order);
	  if (p!=NULL) {
	    p->px=x; p->py=y; p->pz=z;
	  }
  }
}

void ParticleTrail(Sprite* sprite,int length){
  if (sprite!=NULL) {
	  ParticleBatch* p=ParticleBatch::GetParticleBatch(sprite->brush.tex[0],sprite->brush.blend,sprite->order);
	  if (p!=NULL) {
	    p->trail=length;
	  }
  }
}

Entity* PickedEntity(){
	return Pick::PickedEntity();
}

float PickedNX(){
	return Pick::PickedNX();
}

float PickedNY(){
	return Pick::PickedNY();
}

float PickedNZ(){
	return Pick::PickedNZ();
}

Surface* PickedSurface(){
  return Pick::PickedSurface();
}

float PickedTime(){
  return Pick::PickedTime();
}

int PickedTriangle(){
  return Pick::PickedTriangle();
}


float PickedX(){
  return Pick::PickedX();
}

float PickedY(){
  return Pick::PickedY();
}

float PickedZ(){
  return Pick::PickedZ();
}

void PointEntity(Entity* ent,Entity* target_ent,float roll){
  ent->PointEntity(target_ent,roll);
}

void PositionEntity(Entity* ent,float x,float y,float z,int glob){
  ent->PositionEntity(x,y,z, (glob!=0) );
}

// !!! added set/get position from vector
void PositionEntityVector(Entity* ent,const float p[3],int glob){
  ent->PositionEntity(p,(glob!=0));
}
void EntityPositionVector(Entity* ent,float p[3],int glob){
  ent->EntityPosition(p,(glob!=0));
}
// !!! added set/get position and orientation via matrix
void TransformEntity(Entity* ent,const float m[16]){
  ent->TransformEntity(m);
}
void EntityTransform(Entity* ent,float m[16]){
  ent->EntityTransform(m);
}

void PositionMesh(Mesh* mesh,float px,float py,float pz) {
	mesh->PositionMesh(px,py,pz);
}

void PositionTexture(Texture* tex,float u_pos,float v_pos){
	tex->PositionTexture(u_pos,v_pos);
}

float ProjectedX(){
  return Camera::projected_x;
}

float ProjectedY(){
  return Camera::projected_y;
}

float ProjectedZ(){
  return Camera::projected_z;
}


void RenderWorld() {
  Global::RenderWorld();
}

Mesh* RepeatMesh(Mesh* mesh,Entity* parent){
  return mesh->RepeatMesh(parent);
}


void ResetEntity(Entity* ent){
  ent->ResetEntity();
}

void ResetShadow(ShadowObject* shad){
  shad->bVCreated=false;
}

void RotateEntity(Entity* ent,float x,float y,float z,int glob){
  ent->RotateEntity(x,y,z,(glob!=0));
}

void RotateMesh(Mesh* mesh,float pitch,float yaw,float roll){
  mesh->RotateMesh(pitch,yaw,roll);
}

void RotateSprite(Sprite* sprite,float ang){
  sprite->RotateSprite(ang);
}

void RotateTexture(Texture* tex,float ang){
  tex->RotateTexture(ang);
}

void ScaleEntity(Entity* ent,float x,float y,float z,int glob){
  ent->ScaleEntity(x,y,z,(glob!=0));
}

void ScaleMesh(Mesh* mesh,float sx,float sy,float sz){
  mesh->ScaleMesh(sx,sy,sz);
}

void ScaleSprite(Sprite* sprite,float s_x,float s_y){
  sprite->ScaleSprite(s_x,s_y);
}

void ScaleTexture(Texture* tex,float u_scale,float v_scale){
  tex->ScaleTexture(u_scale,v_scale);
}

void SetAnimKey(Entity* ent, float frame, int pos_key, int rot_key, int scale_key){
	ent->SetAnimKey(frame, pos_key, rot_key, scale_key);
}

void SetAnimTime(Entity* ent,float time,int seq) {
  ent->SetAnimTime(time,seq);
}

void SetCubeFace(Texture* tex,int face) {
  tex->cube_face=face;
}

void SetCubeMode(Texture* tex,int mode){
  tex->cube_mode=mode;
}

void ShowEntity(Entity* ent) {
  ent->ShowEntity();
}

void SkinMesh(Mesh* mesh, int surf_no_get, int vid, int bone1, float weight1, int bone2, float weight2, int bone3, float weight3, int bone4, float weight4){
	mesh->SkinMesh(surf_no_get, vid, bone1, weight1, bone2, weight2, bone3, weight3, bone4, weight4);
}

void SpriteViewMode(Sprite* sprite,int mode){
  sprite->SpriteViewMode(mode);
}
void SpriteRenderMode(Sprite* sprite,int mode){
  sprite->SpriteRenderMode(mode);
}

void StencilAlpha(Stencil* stencil, float a){
  stencil->StencilAlpha(a);
}
void StencilClsColor(Stencil* stencil, float r,float g,float b){
  stencil->StencilClsColor(r,g,b);
}
void StencilClsMode(Stencil* stencil,int bClearColor,int bClearDepth){
  stencil->StencilClsMode((bClearColor!=0), (bClearDepth!=0) );
}
void StencilMesh(Stencil* stencil, Mesh* mesh, int mode){
  stencil->StencilMesh(mesh, mode);
}
void StencilMode(Stencil* stencil, int m, int o=1){
  stencil->StencilMode(m, o);
}

float TerrainHeight (Terrain* terr, int x, int z){
  return terr->TerrainHeight (x, z);
}

float TerrainX (Terrain* terr, float x, float y, float z){
  return terr->TerrainX (x, y, z);
}

float TerrainY (Terrain* terr, float x, float y, float z){
  return terr->TerrainY (x, y, z);
}

float TerrainZ (Terrain* terr, float x, float y, float z){
  return terr->TerrainZ (x, y, z);
}

void TextureBlend(Texture* tex, int blend){
  tex->TextureBlend(blend);
}

void TextureCoords(Texture* tex, int coords){
  tex->TextureCoords(coords);
}

int TextureHeight(Texture* tex){
  return tex->height;
}

void TextureFilter(char* match_text,int flags){
  Texture::AddTextureFilter(match_text,flags);
}

const char* TextureName(Texture* tex){
  return tex->TextureName().c_str();
}

int TextureWidth(Texture* tex){
  return tex->width;
}

float TFormedX() {
  return Entity::TFormedX();
}

float TFormedY(){
  return Entity::TFormedY();
}

float TFormedZ(){
  return Entity::TFormedZ();
}

void TFormNormal(float x,float y,float z,Entity* src_ent,Entity* dest_ent){
	Entity::TFormNormal(x,y,z,src_ent,dest_ent);
}

void TFormPoint(float x,float y,float z,Entity* src_ent,Entity* dest_ent){
	Entity::TFormPoint(x,y,z,src_ent,dest_ent);
}

void TFormVector(float x,float y,float z,Entity* src_ent,Entity* dest_ent){
	Entity::TFormVector(x,y,z,src_ent,dest_ent);
}

void TranslateEntity(Entity* ent,float x,float y,float z,int glob){
	ent->TranslateEntity(x,y,z,(glob!=0));
}

int TriangleVertex(Surface* surf,int tri_no,int corner){
	return surf->TriangleVertex(tri_no,corner);
}

void TurnEntity(Entity* ent, float x, float y, float z, int glob){
	if (ent) ent->TurnEntity(x, y, z, (glob!=0));
}

void UpdateNormals(Entity* ent){
  if (ent) {
    Mesh* mesh=dynamic_cast<Mesh*>(ent);
    Terrain* terr=dynamic_cast<Terrain*>(ent);
    Geosphere* geo=dynamic_cast<Geosphere*>(ent);
    if (geo)
      geo->UpdateNormals();
    else if (terr)
      terr->UpdateNormals();
    else if (mesh)
      mesh->UpdateNormals();
  }
}

void MeshUpdateNormals(Mesh* mesh){
  if (mesh)	mesh->UpdateNormals();
}
void TerrainUpdateNormals(Terrain* terr){
  if (terr) terr->UpdateNormals();
}
void GeosphereUpdateNormals(Geosphere* geo,int bPreserve){
  if (geo) geo->UpdateNormals(bPreserve);
}

void UpdateTexCoords(Surface* surf){
	if (surf) surf->UpdateTexCoords();
}

void UpdateWorld(float anim_speed){
	Global::UpdateWorld(anim_speed);
}

void UseStencil(Stencil* stencil){
	if (stencil==0)
		glDisable(GL_STENCIL_TEST);
	else
		stencil->UseStencil();
}

float VectorPitch(float vx,float vy,float vz){
  // !!! changed to float math
	float ang = atan2deg(sqrtf(vx*vx+vz*vz),vy)-90.0;
	if (ang<=0.0001 && ang>=-0.0001)
	  return 0.0f;

	return -ang;
}

float VectorYaw(float vx,float vy,float vz){
	return atan2deg(-vx,vz);
}

float VertexAlpha(Surface* surf,int vid){
	return surf->VertexAlpha(vid);
}

float VertexBlue(Surface* surf,int vid){
	return surf->VertexBlue(vid);
}

void VertexColor(Surface* surf,int vid,float r,float g,float b,float a){
	surf->VertexColor(vid,r,g,b,a);
}

void VertexCoords(Surface* surf,int vid,float x,float y,float z){
	surf->VertexCoords(vid,x,y,z);
}

float VertexGreen(Surface* surf,int vid){
	return surf->VertexGreen(vid);
}

void VertexNormal(Surface* surf,int vid,float nx,float ny,float nz){
	surf->VertexNormal(vid,nx,ny,nz);
}

float VertexNX(Surface* surf,int vid){
	return surf->VertexNX(vid);
}

float VertexNY(Surface* surf,int vid){
	return surf->VertexNY(vid);
}

float VertexNZ(Surface* surf,int vid){
	return surf->VertexNZ(vid);
}

float VertexRed(Surface* surf,int vid){
	return surf->VertexRed(vid);
}

void VertexTexCoords(Surface* surf,int vid,float u,float v,float w,int coord_set){
	surf->VertexTexCoords(vid,u,v,w,coord_set);
}

float VertexU(Surface* surf,int vid,int coord_set){
	return surf->VertexU(vid,coord_set);
}
float VertexV(Surface* surf,int vid,int coord_set){
	return surf->VertexV(vid,coord_set);
}
float VertexW(Surface* surf,int vid,int coord_set){
	return surf->VertexW(vid,coord_set);
}

float VertexX(Surface* surf,int vid){
	return surf->VertexX(vid);
}
float VertexY(Surface* surf,int vid){
	return surf->VertexY(vid);
}
float VertexZ(Surface* surf,int vid){
	return surf->VertexZ(vid);
}

void VoxelSpriteMaterial(VoxelSprite* voxelspr, Material* mat){
	voxelspr->VoxelSpriteMaterial(mat);
}

void Wireframe(int bEnable){
#ifndef GLES2
	if (bEnable!=0)
	  glPolygonMode(GL_FRONT,GL_LINE);
	else
	  glPolygonMode(GL_FRONT,GL_FILL);
#endif

}


//' ***extras***
float EntityScaleX(Entity* ent,int glob){
	return ent->EntityScaleX((glob!=0));
}

float EntityScaleY(Entity* ent,int glob){
	return ent->EntityScaleY((glob!=0));
}

float EntityScaleZ(Entity* ent,int glob){
	return ent->EntityScaleZ((glob!=0));
}

Shader* LoadShader(char* ShaderName, char* VshaderFileName, char* FshaderFileName){
	Shader* shader = Shader::CreateShaderMaterial(ShaderName);
	shader->AddShader(VshaderFileName, GL_VERTEX_SHADER);
	shader->AddShader(FshaderFileName, GL_FRAGMENT_SHADER);
	shader->Link();
	return shader;
}

Shader* CreateShader(char* ShaderName, char* VshaderString, char* FshaderString){
	Shader* shader=Shader::CreateShaderMaterial(ShaderName);
	shader->AddShaderFromString(VshaderString, GL_VERTEX_SHADER);
	shader->AddShaderFromString(FshaderString, GL_FRAGMENT_SHADER);
	shader->Link();
	return shader;
}

Shader* LoadShaderVGF(char* ShaderName, char* VshaderFileName, char* GshaderFileName, char* FshaderFileName){
	Shader* shader = Shader::CreateShaderMaterial(ShaderName);
	shader->AddShader(VshaderFileName, GL_VERTEX_SHADER);
#ifndef GLES2
  shader->AddShader(GshaderFileName, GL_GEOMETRY_SHADER);
#endif
  shader->AddShader(FshaderFileName, GL_FRAGMENT_SHADER);
	shader->Link();
	return shader;
}

Shader* CreateShaderVGF(char* ShaderName, char* VshaderString, char* GshaderString, char* FshaderString){
	Shader* shader = Shader::CreateShaderMaterial(ShaderName);
	shader->AddShaderFromString(VshaderString, GL_VERTEX_SHADER);
#ifndef GLES2
  shader->AddShaderFromString(GshaderString, GL_GEOMETRY_SHADER);
#endif
  shader->AddShaderFromString(FshaderString, GL_FRAGMENT_SHADER);
	shader->Link();
	return shader;
}

void ShadeSurface(Surface* surface, Shader* shader){
	surface->ShaderMat = shader;
}

void ShadeMesh(Mesh* mesh, Shader* shader){
	for(int surfaceindex=1; surfaceindex<=mesh->CountSurfaces(); surfaceindex++){
		Surface* surface = mesh->GetSurface(surfaceindex);
		surface->ShaderMat = shader;
	}
}

void ShadeEntity(Entity* ent, Shader* shader){
	Mesh* mesh = dynamic_cast<Mesh*>(ent);
	Terrain* terrain = dynamic_cast<Terrain*>(ent);

	if (terrain)
		terrain->ShaderMat = shader;
	if (mesh)
		ShadeMesh(mesh, shader);
}

void ShaderTexture(Shader* shader, Texture* tex, char* name, int index,int bIs3D){
	shader->AddSampler(name, index, tex, (bIs3D!=0) );
}

void SetFloat(Shader* shader, char* name, float v1){
	shader->SetFloat(name, v1);
}

void SetFloat2(Shader* shader, char* name, float v1, float v2){
	shader->SetFloat2(name, v1, v2);
}

void SetFloat3(Shader* shader, char* name, float v1, float v2, float v3){
	shader->SetFloat3(name, v1, v2, v3);
}

void SetFloat4(Shader* shader, char* name, float v1, float v2, float v3, float v4){
	shader->SetFloat4(name, v1, v2, v3, v4);
}

void UseFloat(Shader* shader, char* name, float* v1){
	shader->UseFloat(name, v1);
}

void UseFloat2(Shader* shader, char* name, float* v1, float* v2){
	shader->UseFloat2(name, v1, v2);
}

void UseFloat3(Shader* shader, char* name, float* v1, float* v2, float* v3){
	shader->UseFloat3(name, v1, v2, v3);
}

void UseFloat4(Shader* shader, char* name, float* v1, float* v2, float* v3, float* v4){
	shader->UseFloat4(name, v1, v2, v3, v4);
}

void SetInteger(Shader* shader, char* name, int v1){
	shader->SetInteger(name, v1);
}

void SetInteger2(Shader* shader, char* name, int v1, int v2){
	shader->SetInteger2(name, v1, v2);
}

void SetInteger3(Shader* shader, char* name, int v1, int v2, int v3){
	shader->SetInteger3(name, v1, v2, v3);
}

void SetInteger4(Shader* shader, char* name, int v1, int v2, int v3, int v4){
	shader->SetInteger4(name, v1, v2, v3, v4);
}

void UseInteger(Shader* shader, char* name, int* v1){
	shader->UseInteger(name, v1);
}

void UseInteger2(Shader* shader, char* name, int* v1, int* v2){
	shader->UseInteger2(name, v1, v2);
}

void UseInteger3(Shader* shader, char* name, int* v1, int* v2, int* v3){
	shader->UseInteger3(name, v1, v2, v3);
}

void UseInteger4(Shader* shader, char* name, int* v1, int* v2, int* v3, int* v4){
	shader->UseInteger4(name, v1, v2, v3, v4);
}

void UseSurface(Shader* shader, char* name, Surface* surf, int vbo){
	shader->UseSurface(name, surf, vbo);
}

void UseMatrix(Shader* shader, char* name, int mode){
	shader->UseMatrix(name, mode);
}

void UseEntity(Shader* shader, char* name, Entity* ent, int mode){
	shader->UseEntity(name, ent, mode);
}

void ShaderFunction(Shader* shader, void (*EnableFunction)(void), void (*DisableFunction)(void)){
	shader->UseFunction(EnableFunction, DisableFunction);
}

Material* LoadMaterial(char* filename,int flags, int frame_width,int frame_height,int first_frame,int frame_count){
	return Material::LoadMaterial(filename, flags, frame_width, frame_height, first_frame, frame_count);
}

void ShaderMaterial(Shader* shader, Material* tex, char* name, int index){
	shader->AddSampler(name, index, tex, 1);
}

void AmbientShader(Shader* shader){
	Global::ambient_shader=shader;
}

int GetShaderProgram(Shader* shader){
	return shader->GetProgram();
}

OcTree* CreateOcTree(float w, float h, float d, Entity* parent_ent){
	return OcTree::CreateOcTree(w, h, d, parent_ent);
}

void OctreeBlock(OcTree* octree, Mesh* mesh, int level, float X, float Y, float Z, float Near, float Far){
	octree->OctreeBlock(mesh, level, X, Y, Z, Near, Far);
}

void OctreeMesh(OcTree* octree, Mesh* mesh, int level, float X, float Y, float Z, float Near, float Far){
	octree->OctreeMesh(mesh, level, X, Y, Z, Near, Far);
}

Fluid* CreateFluid(){
	return Fluid::CreateFluid();
}

ParticleEmitter* CreateParticleEmitter(Entity* particle, Entity* parent_ent){
	return ParticleEmitter::CreateParticleEmitter(particle, parent_ent);
}

Action* ActMoveBy(Entity* ent, float a, float b, float c, float rate){
	return Action::AddAction(ent, ACT_MOVEBY, a, b, c, rate);
}

Action* ActTurnBy(Entity* ent, float a, float b, float c, float rate){
	return Action::AddAction(ent, ACT_TURNBY, a, b, c, rate);
}

Action* ActVector(Entity* ent, float a, float b, float c){
	return Action::AddAction(ent, ACT_VECTOR, a, b, c, 0);
}

Action* ActMoveTo(Entity* ent, float a, float b, float c, float rate){
	return Action::AddAction(ent, ACT_MOVETO, a, b, c, rate);
}

Action* ActTurnTo(Entity* ent, float a, float b, float c, float rate){
	return Action::AddAction(ent, ACT_TURNTO, a, b, c, rate);
}

Action* ActScaleTo(Entity* ent, float a, float b, float c, float rate){
	return Action::AddAction(ent, ACT_SCALETO, a, b, c, rate);
}

Action* ActFadeTo(Entity* ent, float a, float rate){
	return Action::AddAction(ent, ACT_FADETO, a, 0, 0, rate);
}

Action* ActTintTo(Entity* ent, float a, float b, float c, float rate){
	return Action::AddAction(ent, ACT_TINTTO, a, b, c, rate);
}

Action* ActTrackByPoint(Entity* ent, Entity* target, float a, float b, float c, float rate){
	return Action::AddAction(ent, ACT_TRACK_BY_POINT, target, a, b, c, rate);
}

Action* ActTrackByDistance(Entity* ent, Entity* target, float a, float rate){
	return Action::AddAction(ent, ACT_TRACK_BY_DISTANCE, target, a, 0, 0, rate);
}

Action* ActNewtonian(Entity* ent, float rate){
	return Action::AddAction(ent, ACT_NEWTONIAN, ent->mat.grid[3][0], ent->mat.grid[3][1], -ent->mat.grid[3][2], rate);
}

void AppendAction(Action* act1, Action* act2){
	act1->AppendAction(act2);
}

void FreeAction(Action* act){
	act->FreeAction();
}

void DepthBufferToTex( Texture* tex, Camera* cam){
	tex->DepthBufferToTex(cam);
}

PostFX* CreatePostFX(Camera* cam, int passes) {
	return PostFX::CreatePostFX(cam, passes);
}
void PostFXAddRenderTarget(PostFX* fx, int pass_no, int numColBufs, int bDepth, int format, float scale){
	fx->PostFXAddRenderTarget(pass_no, numColBufs, (bDepth!=0), format, scale);
}
void PostFXShader(PostFX* fx, int pass_no, Shader* shader){
	fx->PostFXShader(pass_no, shader);
}
void PostFXShaderPass(PostFX* fx, int pass_no, char* name, int v){
	fx->PostFXShaderPass(pass_no, name, v);
}
void PostFXBuffer(PostFX* fx, int pass_no, int source_pass, int index, int slot){
	fx->PostFXBuffer(pass_no, source_pass, index, slot);
}
void PostFXTexture(PostFX* fx, int pass_no, Texture* tex, int slot, int frame){
	fx->PostFXTexture(pass_no, tex, slot, frame);
}
void PostFXFunction(PostFX* fx, int pass_no, void (*PassFunction)(int pass_no)){
	fx->PostFXFunction(pass_no, PassFunction);
}

float* CameraProjMatrix(Camera* cam){
  return cam->CameraProjMatrix();
}
float* EntityMatrix(Entity* ent){
  return ent->EntityMatrix();
}

// !!! new
void ActSetUserdata(Action* act,void* ud){
  act->SetUserdata(ud);
}
void* ActGetUserdata(Action* act){
  return act->GetUserdata();
}
void BoneSetUserdata(Bone* bone,void* ud){
  bone->SetUserdata(ud);
}
void* BoneGetUserdata(Bone* bone){
  return bone->GetUserdata();
}
void BrushSetUserdata(Brush* brush,void* ud){
  brush->SetUserdata(ud);
}
void* BrushGetUserdata(Brush* brush){
  return brush->GetUserdata();
}
void ConstraintSetUserdata(Constraint* con,void* ud){
  con->SetUserdata(ud);
}
void* ConstraintGetUserdata(Constraint* con){
  return con->GetUserdata();
}
void CameraSetUserdata(Camera* cam,void* ud){
  cam->SetUserdata(ud);
}
void* CameraGetUserdata(Camera* cam){
  return cam->GetUserdata();
}
void GeosphereSetUserdata(Geosphere* geo,void* ud){
  geo->SetUserdata(ud);
}
void* GeosphereGetUserdata(Geosphere* geo){
  return geo->GetUserdata();
}
void BlobSetUserdata(Blob* blob,void* ud){
  blob->SetUserdata(ud);
}
void* BlobGetUserdata(Blob* blob){
  return blob->GetUserdata();
}
void FluidSetUserdata(Fluid* fluid,void* ud){
  fluid->SetUserdata(ud);
}
void* FluidGetUserdata(Fluid* fluid){
  return fluid->GetUserdata();
}
void LightSetUserdata(Light* light,void* ud){
  light->SetUserdata(ud);
}
void* LightGetUserdata(Light* light){
  return light->GetUserdata();
}
void MeshSetUserdata(Mesh* mesh,void* ud){
  mesh->SetUserdata(ud);
}
void* MeshGetUserdata(Mesh* mesh){
  return mesh->GetUserdata();
}
void OctreeSetUserdata(OcTree* octree,void* ud){
  octree->SetUserdata(ud);
}
void* OctreeGetUserdata(OcTree* octree){
  return octree->GetUserdata();
}
void EmitterSetUserdata(ParticleEmitter* emitter,void* ud){
  emitter->SetUserdata(ud);
}
void* EmitterGetUserdata(ParticleEmitter* emitter){
  return emitter->GetUserdata();
}
void RigidBodySetUserdata(RigidBody* body,void* ud){
  body->SetUserdata(ud);
}
void* RigidBodyGetUserdata(RigidBody* body){
  return body->GetUserdata();
}
void PivotSetUserdata(Pivot* pivot,void* ud){
  pivot->SetUserdata(ud);
}
void* PivotGetUserdata(Pivot* pivot){
  return pivot->GetUserdata();
}
void PostFXSetUserdata(PostFX* postfx,void* ud){
  postfx->SetUserdata(ud);
}
void* PostFXGetUserdata(PostFX* postfx){
  return postfx->GetUserdata();
}
void ShaderSetUserdata(Shader* shader,void* ud){
  shader->SetUserdata(ud);
}
void* ShaderGetUserdata(Shader* shader){
  return shader->GetUserdata();
}
void ShadowObjectSetUserdata(ShadowObject* shadowobj,void* ud){
  shadowobj->SetUserdata(ud);
}
void* ShadowObjectGetUserdata(ShadowObject* shadowobj){
  return shadowobj->GetUserdata();
}
void SpriteSetUserdata(Sprite* sprite,void* ud){
  sprite->SetUserdata(ud);
}
void* SpriteGetUserdata(Sprite* sprite){
  return sprite->GetUserdata();
}
void StencilSetUserdata(Stencil* stencil,void* ud){
  stencil->SetUserdata(ud);
}
void* StencilGetUserdata(Stencil* stencil){
  return stencil->GetUserdata();
}
void SurfaceSetUserdata(Surface* sur,void* ud){
  sur->SetUserdata(ud);
}
void* SurfaceGetUserdata(Surface* sur){
  return sur->GetUserdata();
}
void TerrainSetUserdata(Terrain* terrain,void* ud){
  terrain->SetUserdata(ud);
}
void* TerrainGetUserdata(Terrain* terrain){
  return terrain->GetUserdata();
}
void TextureSetUserdata(Texture* tex,void* ud){
  tex->SetUserdata(ud);
}
void* TextureGetUserdata(Texture* tex){
  return tex->GetUserdata();
}
void VoxelSpriteSetUserdata(VoxelSprite* voxelsprite,void* ud){
  voxelsprite->SetUserdata(ud);
}
void* VoxelSpriteGetUserdata(VoxelSprite* voxelsprite){
  return voxelsprite->GetUserdata();
}

void PreUpdateCalback (void (*cb)(void* userdata),void* ud){
  Global::PreUpdateCalback(cb,ud);
}
void PostUpdateCalback(void (*cb)(void* userdata),void* ud){
  Global::PostUpdateCalback(cb,ud);
}
void PreRenderCalback (void (*cb)(void* userdata),void* ud){
  Global::PreRenderCalback(cb,ud);
}
void PostRenderCalback(void (*cb)(void* userdata),void* ud){
  Global::PostRenderCalback(cb,ud);
}

int EntityHidden(Entity* ent) {
  return ent->Hidden();
}
void EntityColor2(Entity* ent,float red,float green,float blue,float alpha,int bRecursive) {
  ent->EntityColor(red,green,blue,alpha,(bRecursive!=0));
}

void AlignToVector(Entity* ent,float x,float y,float z, int axis, float rate){
  ent->AlignToVector(x,y,z,axis,rate);
}

void EntityRed(Entity* ent,float r,int bRecursive){
  ent->EntityRed(r,(bRecursive!=0));
}
void EntityGreen(Entity* ent,float g,int bRecursive){
  ent->EntityGreen(g,(bRecursive!=0));
}
void EntityBlue(Entity* ent,float b,int bRecursive){
  ent->EntityBlue(b,(bRecursive!=0));
}

/*void SetParameter1S(Shader* material, char* name, float v1){
	material->SetParameter1S(name, v1);
}

void SetParameter2S(Shader* material, char* name, float v1, float v2){
	material->SetParameter2S(name, v1, v2);
}

void SetParameter3S(Shader* material, char* name, float v1, float v2, float v3){
	material->SetParameter3S(name, v1, v2, v3);
}

void SetParameter4S(Shader* material, char* name, float v1, float v2, float v3, float v4){
	material->SetParameter4S(name, v1, v2, v3, v4);
}

void SetParameter1I(Shader* material, char* name, int v1){
	material->SetParameter1I(name, v1);
}

void SetParameter2I(Shader* material, char* name, int v1, int v2){
	material->SetParameter2I(name, v1, v2);
}

void SetParameter3I(Shader* material, char* name, int v1, int v2, int v3){
	material->SetParameter3I(name, v1, v2, v3);
}

void SetParameter4I(Shader* material, char* name, int v1, int v2, int v3, int v4){
	material->SetParameter4I(name, v1, v2, v3, v4);
}

void SetVector1I(Shader* material, char* name, int* v1){
	material->SetVector1I(name, v1);
}

void SetVector2I(Shader* material, char* name, int* v1){
	material->SetVector2I(name, v1);
}

void SetVector3I(Shader* material, char* name, int* v1){
	material->SetVector3I(name, v1);
}

void SetVector4I(Shader* material, char* name, int* v1){
	material->SetVector4I(name, v1);
}

void SetParameter1F(Shader* material, char* name, float v1){
	material->TurnOn();
	material->SetParameter1F(name, v1);
	material->TurnOff();
}

void SetParameter2F(Shader* material, char* name, float v1, float v2){
	material->SetParameter2F(name, v1, v2);
}

void SetParameter3F(Shader* material, char* name, float v1, float v2, float v3){
	material->SetParameter3F(name, v1, v2, v3);
}

void SetParameter4F(Shader* material, char* name, float v1, float v2, float v3, float v4){
	material->SetParameter4F(name, v1, v2, v3, v4);
}

void SetVector1F(Shader* material, char* name, float* v1){
	material->SetVector1F(name, v1);
}

void SetVector2F(Shader* material, char* name, float* v1){
	material->SetVector2F(name, v1);
}

void SetVector3F(Shader* material, char* name, float* v1){
	material->SetVector3F(name, v1);
}

void SetVector4F(Shader* material, char* name, float* v1){
	material->SetVector4F(name, v1);
}

void SetMatrix2F(Shader* material, char* name, float* m){
	material->SetMatrix2F(name, m);
}

void SetMatrix3F(Shader* material, char* name, float* m){
	material->SetMatrix3F(name, m);
}

void SetMatrix4F(Shader* material, char* name, float* m){
	material->SetMatrix4F(name, m);
}

void SetParameter1D(Shader* material, char* name, double v1){
	material->SetParameter1S(name, v1);
}

void SetParameter2D(Shader* material, char* name, double v1, double v2){
	material->SetParameter2S(name, v1, v2);
}

void SetParameter3D(Shader* material, char* name, double v1, double v2, double v3){
	material->SetParameter3S(name, v1, v2, v3);
}

void SetParameter4D(Shader* material, char* name, double v1, double v2, double v3, double v4){
	material->SetParameter4S(name, v1, v2, v3, v4);
}
*/

//' ***todo***

/*
Function LightMesh(mesh:TMesh,red#,green#,blue#,range#=0,light_x#=0,light_y#=0,light_z#=0)
End Function
Function SetAnimKey(ent:TEntity,frame,pos_key=True,rot_key=True,scale_key=True)
End Function
*/

#ifdef __cplusplus
} /* extern "C" */
#endif


