#ifdef EMSCRIPTEN
 #include <GLES2/gl2.h>
 #define GLES2
#else
 /// windows ,linux ,mac
 #include "GLee.h"
#endif

#include "shadow.h"
#include "light.h"
#include "mesh.h"
#include "global.h"

float ShadowObject::ShadowRed   =0;
float ShadowObject::ShadowGreen =0;
float ShadowObject::ShadowBlue  =0;
float ShadowObject::ShadowAlpha =.5;

list<ShadowObject*> ShadowObject::shadow_list;

float ShadowObject::light_x, ShadowObject::light_y, ShadowObject::light_z;

/* !!! not used or changed at all
  int ShadowObject::top_caps=1;
 int ShadowObject::parallel=0; */

float ShadowObject::VolumeLength=1000;

int ShadowObject::midStencilVal;

void ShadowObject::FreeShadow() {
	shadow_list.remove(this);
	ShadowMesh->FreeEntity();
	vector<ShadowTriangle*>::iterator it;
	for(it=ShadowTriangles.begin();it!=ShadowTriangles.end();it++){
		delete *it;
	}
	if (shadow_list.size()==0)
    Global::Shadows_enabled=0;
}

void ShadowObject::SetShadowColor(int R, int G, int B, int A){
	ShadowRed   = R;
	ShadowGreen = G;
	ShadowBlue  = B;
	ShadowAlpha = A;
}

void ShadowObject::ShadowInit(){
  int StencilBits;
  glGetIntegerv(GL_STENCIL_BITS, &StencilBits);
  ShadowObject::midStencilVal = (StencilBits - 1)^2;
  glClearStencil(ShadowObject::midStencilVal);
}

ShadowObject* ShadowObject::Create(Mesh* mesh, bool bStatic, Mesh* mesh2){
  ShadowObject* S = new ShadowObject;
  S->ShadowMesh = Mesh::CreateMesh();
  Global::root_ent->child_list.remove(S->ShadowMesh);

  S->ShadowMesh->EntityBlend(0);
  S->ShadowMesh->EntityAlpha(0.2);
  S->ShadowMesh->EntityColor(1, 0, 0);
  S->ShadowMesh->EntityFX(17); // 16 + 1 FX_DOUBLESIDED | FX_FULLBRIGHT

  S->ShadowVolume = S->ShadowMesh->CreateSurface();
  S->Parent1 = mesh;
  S->Parent2 = mesh2;
  S->bStatic = bStatic;
  S->bVCreated=0;

	shadow_list.push_back(S);
	S->Init();
	return S;
}

void ShadowObject::SetUserdata(void* ud){
  userdata=ud;
}
void* ShadowObject::GetUserdata(void){
  return userdata;
}

void ShadowObject::RemoveShadowfromMesh(Mesh* parent) {
	list<ShadowObject*>::iterator it;
	for(it=shadow_list.begin();it!=shadow_list.end();it++){
		ShadowObject* S=*it;
		if (S->Parent1 == parent){
			shadow_list.remove(S);
			return;
		}
	}
}

void ShadowObject::Update(Camera* Cam){
  vector<Light*>::iterator light_it;
  Light* light;
  // !!! save current cls mode
  bool old_cls_color;
  bool old_cls_zbuffer;
  Cam->CameraGetClsMode(old_cls_color, old_cls_zbuffer);

  for(light_it=Light::light_list.begin();light_it!=Light::light_list.end();++light_it){
    light=*light_it;
    if (light->hide==true || light->cast_shadow==false)
      continue;

    light_x = light->EntityX(true); // !!! * (1 + parallel * 1);
    light_y = light->EntityY(true); // !!! * (1 + parallel * 1);
    light_z = light->EntityZ(true); // !!! * (1 + parallel * 1);
	  // overwrite it
    Cam->CameraClsMode(false,false);
		Cam->Update();

		list<ShadowObject*>::iterator shadowobject_it;
		ShadowObject* shadowobject;
		for(shadowobject_it=shadow_list.begin();shadowobject_it!=shadow_list.end();shadowobject_it++){

			shadowobject = *shadowobject_it;

			if (shadowobject->Parent1->hide==false && light->EntityDistance(shadowobject->Parent1) < 1/light->range) {
				ShadowObject::VolumeLength = (Cam->range_far - shadowobject->Parent1->EntityDistance(Cam)) / (shadowobject->Parent1->EntityDistance(light) + fabsf(shadowobject->Parent1->cull_radius));
				shadowobject->ShadowMesh->reset_bounds = true;
				shadowobject->ShadowMesh->GetBounds();
				shadowobject->UpdateCaster();
				shadowobject->bRender = true;
			} else {
				shadowobject->bRender = false;
			}
		}
    // !!! Cam->CameraClsMode(false,false);
    ShadowRenderWorldZFail();
	}
  // !!! restore cls mode
	Cam->CameraClsMode(old_cls_color, old_cls_zbuffer);
}

void ShadowObject::RenderVolume(){
	glEnable(GL_POLYGON_OFFSET_FILL);
	glPolygonOffset(0.0, 4.0);
  ShadowObject* shadowobject;
	list<ShadowObject*>::iterator it;
	for(it=shadow_list.begin();it!=shadow_list.end();it++){
		shadowobject=*it;
		if (shadowobject->bRender) {
      shadowobject->ShadowMesh->UpdateShadow();
    }
	}
	glDisable(GL_POLYGON_OFFSET_FILL);
}

void ShadowObject::UpdateAnim(){
	Surface* surface;
	Surface* animsurface;
  Mesh* Parent=Parent2;
  if (Parent==NULL)
    Parent=Parent1;
	list<Surface*>::iterator surface_it = Parent->surf_list.begin();
	list<Surface*>::iterator animsurface_it = Parent->anim_surf_list.begin();
	vector<ShadowTriangle*>::iterator shadowtriangle_it = ShadowTriangles.begin();
  ShadowTriangle* shadowtriangle;
	int cnt;
	int cnt_surf = Parent->CountSurfaces();
	for (int s = 1; s<= cnt_surf;s++){
		surface = *surface_it;
		animsurface = *animsurface_it;
		cnt = surface->CountTriangles() - 1;

		for (int v = 0;v<=cnt;v++){
			shadowtriangle = *shadowtriangle_it;
			int vert0 = surface->TriangleVertex(shadowtriangle->tris, 0);
			int vert1 = surface->TriangleVertex(shadowtriangle->tris, 1);
			int vert2 = surface->TriangleVertex(shadowtriangle->tris, 2);
			shadowtriangle->v1x = animsurface->VertexX(vert0);
			shadowtriangle->v1y = animsurface->VertexY(vert0);
			shadowtriangle->v1z = animsurface->VertexZ(vert0);

			shadowtriangle->v2x = animsurface->VertexX(vert1);
			shadowtriangle->v2y = animsurface->VertexY(vert1);
			shadowtriangle->v2z = animsurface->VertexZ(vert1);

			shadowtriangle->v3x = animsurface->VertexX(vert2);
			shadowtriangle->v3y = animsurface->VertexY(vert2);
			shadowtriangle->v3z = animsurface->VertexZ(vert2);
			shadowtriangle_it++;
		}
		surface_it++;
		animsurface_it++;
	}
}

void ShadowObject::Init(){
	int nTriangles = -1;
	ShadowObject::cnt_tris = -1;
	Surface* surface;
	// !!! new
  Mesh* Parent=ShadowObject::Parent2;
  if (Parent==NULL)
    Parent=ShadowObject::Parent1;

	int nSurfaces = Parent->CountSurfaces();
	for (int surfaceindex = 1; surfaceindex <= nSurfaces; surfaceindex++){
		surface = Parent->GetSurface(surfaceindex);
		nTriangles = surface->CountTriangles() - 1;

		for (int triangleindex = 0;triangleindex<=nTriangles;triangleindex++){
			ShadowObject::cnt_tris += 1;
			ShadowTriangle* shadowtriangle = new ShadowTriangle;
			ShadowTriangles.push_back(shadowtriangle);
			shadowtriangle->tris = triangleindex;
			int vert0 = surface->TriangleVertex(triangleindex, 0);
			int vert1 = surface->TriangleVertex(triangleindex, 1);
			int vert2 = surface->TriangleVertex(triangleindex, 2);

			shadowtriangle->v1x = surface->VertexX(vert0);
			shadowtriangle->v1y = surface->VertexY(vert0);
			shadowtriangle->v1z = surface->VertexZ(vert0);

			shadowtriangle->v2x = surface->VertexX(vert1);
			shadowtriangle->v2y = surface->VertexY(vert1);
			shadowtriangle->v2z = surface->VertexZ(vert1);

			shadowtriangle->v3x = surface->VertexX(vert2);
			shadowtriangle->v3y = surface->VertexY(vert2);
			shadowtriangle->v3z = surface->VertexZ(vert2);

			shadowtriangle->ta = -1;
			shadowtriangle->tb = -1;
			shadowtriangle->tc = -1;
		}
	}

	float v1a_x, v1a_y, v1a_z,
	      v1b_x, v1b_y, v1b_z,
	      v1c_x, v1c_y, v1c_z;

	bool v1a_v2a, v1a_v2b, v1a_v2c,
	     v1b_v2a, v1b_v2b, v1b_v2c,
	     v1c_v2a, v1c_v2b, v1c_v2c;

	for (int a = 0; a <= ShadowObject::cnt_tris; a++) {
		ShadowTriangle* at = ShadowTriangles[a];
		v1a_x = at->v1x; v1a_y = at->v1y;	v1a_z = at->v1z;
		v1b_x = at->v2x; v1b_y = at->v2y;	v1b_z = at->v2z;
		v1c_x = at->v3x; v1c_y = at->v3y;	v1c_z = at->v3z;

		for (int b = a + 1;b <= ShadowObject::cnt_tris; b++){
			ShadowTriangle* bt = ShadowTriangles[b];
			v1a_v2a = (v1a_x == bt->v1x && v1a_y == bt->v1y && v1a_z == bt->v1z);
			v1a_v2b = (v1a_x == bt->v2x && v1a_y == bt->v2y && v1a_z == bt->v2z);
			v1a_v2c = (v1a_x == bt->v3x && v1a_y == bt->v3y && v1a_z == bt->v3z);

			v1b_v2a = (v1b_x == bt->v1x && v1b_y == bt->v1y && v1b_z == bt->v1z);
			v1b_v2b = (v1b_x == bt->v2x && v1b_y == bt->v2y && v1b_z == bt->v2z);
			v1b_v2c = (v1b_x == bt->v3x && v1b_y == bt->v3y && v1b_z == bt->v3z);

			v1c_v2a = (v1c_x == bt->v1x && v1c_y == bt->v1y && v1c_z == bt->v1z);
			v1c_v2b = (v1c_x == bt->v2x && v1c_y == bt->v2y && v1c_z == bt->v2z);
			v1c_v2c = (v1c_x == bt->v3x && v1c_y == bt->v3y && v1c_z == bt->v3z);

			if (bt->ta==-1){
				if      (v1a_v2b!=0 && v1b_v2a!=0 && at->ta==-1){
					at->ta = b;
					bt->ta = a;
				}
				else if (v1b_v2b!=0 && v1c_v2a!=0 && at->tb==-1){
					at->tb = b;
					bt->ta = a;
				}
				else if (v1c_v2b!=0 && v1a_v2a!=0 && at->ta==-1){
					at->tc = b;
					bt->ta = a;
				}
			}
			if (bt->tb==-1){
				if (v1a_v2c!=0 && v1b_v2b!=0 && at->ta==-1){
					at->ta = b;
					bt->tb = a;
				}
				else if (v1b_v2c!=0 && v1c_v2b!=0 && at->tb==-1){
					at->tb = b;
					bt->tb = a;
				}
				else if (v1c_v2c!=0 && v1a_v2b!=0 && at->tc==-1){
					at->tc = b;
					bt->tb = a;
				}
			}
			if (bt->tc==-1){
				if (v1a_v2a!=0 && v1b_v2c!=0 && at->ta==-1){
					at->ta = b;
					bt->tc = a;
				}
				else if (v1b_v2a!=0 && v1c_v2c!=0 && at->tb==-1){
					at->tb = b;
					bt->tc = a;
				}
				else if (v1c_v2a!=0 && v1a_v2c!=0 && at->tc==-1){
					at->tc = b;
					bt->tc = a;
				}
			}
		}
	}
	InitShadow();
}

void ShadowObject::InitShadow(){
	if (Global::Shadows_enabled==0){
	  Global::Shadows_enabled=true;
	  ShadowInit();
	}
	ShadowTriangle* shadowtriangle;
	for (int v = 0;v<=cnt_tris;v++){
		shadowtriangle = ShadowTriangles[v];
		if (shadowtriangle->ta > -1)
		  shadowtriangle->shadowtriangle_a = ShadowTriangles[shadowtriangle->ta];
		if (shadowtriangle->tb > -1)
		  shadowtriangle->shadowtriangle_b = ShadowTriangles[shadowtriangle->tb];
		if (shadowtriangle->tc > -1)
		  shadowtriangle->shadowtriangle_c = ShadowTriangles[shadowtriangle->tc];
	}
}

void ShadowObject::UpdateCaster() {
	if (bStatic && bVCreated)
    return;

	bVCreated = true;

	if (ShadowVolume != 0) {
    ShadowVolume->ClearSurface();
  }

  Mesh* Parent=Parent2;
  if (Parent==NULL)
    Parent=Parent1;

	if (Parent->anim != 0) {
    UpdateAnim();
  }

	float e0x, e0y, e0z, e1x, e1y, e1z;
	float normlight_x, normlight_y, normlight_z;
	float r1x, r1y, r1z, r2x, r2y, r2z, r3x, r3y, r3z;
	int vert1, vert2, vert3, vert4;
	char check1=0, check2=0, check3=0;

	ShadowTriangle* shadowtriangle;
	vector<ShadowTriangle*>::iterator shadowtriangle_it;
	for(shadowtriangle_it=ShadowTriangles.begin();shadowtriangle_it!=ShadowTriangles.end();shadowtriangle_it++){
		shadowtriangle= *shadowtriangle_it;

		shadowtriangle->tf_v1x = shadowtriangle->v1x;
		shadowtriangle->tf_v1y = shadowtriangle->v1y;
		shadowtriangle->tf_v1z = shadowtriangle->v1z;
		Parent->mat.TransformVec(shadowtriangle->tf_v1x, shadowtriangle->tf_v1y, shadowtriangle->tf_v1z, 1);

		shadowtriangle->tf_v2x = shadowtriangle->v2x;
		shadowtriangle->tf_v2y = shadowtriangle->v2y;
		shadowtriangle->tf_v2z = shadowtriangle->v2z;
		Parent->mat.TransformVec(shadowtriangle->tf_v2x, shadowtriangle->tf_v2y, shadowtriangle->tf_v2z, 1);

		shadowtriangle->tf_v3x = shadowtriangle->v3x;
		shadowtriangle->tf_v3y = shadowtriangle->v3y;
		shadowtriangle->tf_v3z = shadowtriangle->v3z;
		Parent->mat.TransformVec(shadowtriangle->tf_v3x, shadowtriangle->tf_v3y, shadowtriangle->tf_v3z, 1);

		e0x = shadowtriangle->tf_v3x - shadowtriangle->tf_v2x;
		e0y = shadowtriangle->tf_v3y - shadowtriangle->tf_v2y;
		e0z = shadowtriangle->tf_v3z - shadowtriangle->tf_v2z;

		e1x = shadowtriangle->tf_v2x - shadowtriangle->tf_v1x;
		e1y = shadowtriangle->tf_v2y - shadowtriangle->tf_v1y;
		e1z = shadowtriangle->tf_v2z - shadowtriangle->tf_v1z;

		normlight_x = (shadowtriangle->tf_v1x - light_x)*(e0y*e1z - e0z*e1y);
		normlight_y = (shadowtriangle->tf_v1y - light_y)*(e0z*e1x - e0x*e1z);
		normlight_z = (shadowtriangle->tf_v1z - light_z)*(e0x*e1y - e0y*e1x);

		shadowtriangle->cull = (normlight_x + normlight_y + normlight_z > 0);
	}

	for(shadowtriangle_it=ShadowTriangles.begin();shadowtriangle_it!=ShadowTriangles.end();shadowtriangle_it++){
		shadowtriangle = *shadowtriangle_it;
		r1x = shadowtriangle->tf_v1x + (shadowtriangle->tf_v1x - light_x)*VolumeLength;
		r1y = shadowtriangle->tf_v1y + (shadowtriangle->tf_v1y - light_y)*VolumeLength;
		r1z = shadowtriangle->tf_v1z + (shadowtriangle->tf_v1z - light_z)*VolumeLength;

		r2x = shadowtriangle->tf_v2x + (shadowtriangle->tf_v2x - light_x)*VolumeLength;
		r2y = shadowtriangle->tf_v2y + (shadowtriangle->tf_v2y - light_y)*VolumeLength;
		r2z = shadowtriangle->tf_v2z + (shadowtriangle->tf_v2z - light_z)*VolumeLength;

		r3x = shadowtriangle->tf_v3x + (shadowtriangle->tf_v3x - light_x)*VolumeLength;
		r3y = shadowtriangle->tf_v3y + (shadowtriangle->tf_v3y - light_y)*VolumeLength;
		r3z = shadowtriangle->tf_v3z + (shadowtriangle->tf_v3z - light_z)*VolumeLength;

		if (shadowtriangle->cull == 0) {

			if (shadowtriangle->ta > -1) check1 = shadowtriangle->shadowtriangle_a->cull; else check1 = 1;
      if (shadowtriangle->tb > -1) check2 = shadowtriangle->shadowtriangle_b->cull; else check2 = 1;
			if (shadowtriangle->tc > -1) check3 = shadowtriangle->shadowtriangle_c->cull; else check3 = 1;

			if (check1!=0){
				vert1 = ShadowVolume->AddVertex(shadowtriangle->tf_v1x , shadowtriangle->tf_v1y , shadowtriangle->tf_v1z);
				vert2 = ShadowVolume->AddVertex(shadowtriangle->tf_v2x , shadowtriangle->tf_v2y , shadowtriangle->tf_v2z);
				vert3 = ShadowVolume->AddVertex(r1x, r1y, r1z);
				vert4 = ShadowVolume->AddVertex(r2x, r2y, r2z);
				ShadowVolume->AddTriangle(vert1, vert3, vert4);
				ShadowVolume->AddTriangle(vert1, vert4, vert2);
			}
			if (check2!=0){
				vert1 = ShadowVolume->AddVertex(shadowtriangle->tf_v2x , shadowtriangle->tf_v2y , shadowtriangle->tf_v2z);
				vert2 = ShadowVolume->AddVertex(shadowtriangle->tf_v3x , shadowtriangle->tf_v3y , shadowtriangle->tf_v3z);
				vert3 = ShadowVolume->AddVertex(r2x, r2y, r2z);
				vert4 = ShadowVolume->AddVertex(r3x, r3y, r3z);
				ShadowVolume->AddTriangle(vert1, vert3, vert4);
				ShadowVolume->AddTriangle(vert1, vert4, vert2);
			}
			if (check3!=0){
				vert1 = ShadowVolume->AddVertex(shadowtriangle->tf_v1x, shadowtriangle->tf_v1y, shadowtriangle->tf_v1z);
				vert2 = ShadowVolume->AddVertex(shadowtriangle->tf_v3x, shadowtriangle->tf_v3y, shadowtriangle->tf_v3z);
				vert3 = ShadowVolume->AddVertex(r1x, r1y, r1z);
				vert4 = ShadowVolume->AddVertex(r3x, r3y, r3z);
				ShadowVolume->AddTriangle(vert1, vert4, vert3);
				ShadowVolume->AddTriangle(vert1, vert2, vert4);
			}

			// !!! if (ShadowObject::top_caps!=0) {
			  vert1 = ShadowVolume->AddVertex(shadowtriangle->tf_v1x, shadowtriangle->tf_v1y, shadowtriangle->tf_v1z);
			  vert2 = ShadowVolume->AddVertex(shadowtriangle->tf_v2x, shadowtriangle->tf_v2y, shadowtriangle->tf_v2z);
			  vert3 = ShadowVolume->AddVertex(shadowtriangle->tf_v3x, shadowtriangle->tf_v3y, shadowtriangle->tf_v3z);
			  ShadowVolume->AddTriangle(vert1, vert2, vert3);
			// }

			// !!! if (ShadowObject::parallel==0){
			  vert1 = ShadowVolume->AddVertex(r1x, r1y, r1z);
			  vert2 = ShadowVolume->AddVertex(r2x, r2y, r2z);
			  vert3 = ShadowVolume->AddVertex(r3x, r3y, r3z);
			  ShadowVolume->AddTriangle(vert1, vert3, vert2);
			// }
		}
	}
}

void ShadowObject::ShadowRenderWorldZFail(){
	glEnable(GL_STENCIL_TEST);
	/*Pass 1 Ambient Pass First clear the stencil Buffer */
	glClearStencil(midStencilVal);
	glClear(GL_STENCIL_BUFFER_BIT);

	glDepthMask(GL_FALSE);
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
	// we wanted the first ambient pass above in the Color Buffer
	// but we don't want the volumes in
	// there ..removing the above line is fun though :P
	glEnable(GL_CULL_FACE);
	glEnable(GL_STENCIL_TEST);

	// Render the shadow volume And increment the stencil every where a front
	// facing polygon is rendered.

	// Incrementing the stencil Buffer when back face depth fails
	glStencilFunc(GL_ALWAYS, ShadowObject::midStencilVal, 0xffffffff); // ~0 is like 0xFFFFFFFF Or something :P
	glStencilOp(GL_KEEP, GL_INCR_WRAP, GL_KEEP); // incrementing on the depth fail
	glCullFace(GL_BACK); // cull front facing polys For this pass

	RenderVolume();

	// Render the shadow volume And decrement the stencil every where a back
	// facing polygon is rendered.
	glStencilOp(GL_KEEP, GL_DECR_WRAP, GL_KEEP); // decrementing on the depth fail
	glCullFace(GL_FRONT); // And now culling back facing polys

	RenderVolume();

	// When done, set the states back To something more typical.
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
#ifndef GLES2
	glEnable(GL_LIGHTING);
#endif
	glDepthMask(GL_TRUE);
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

	//
	// Render the lit part...
	//
	glStencilFunc(GL_EQUAL, ShadowObject::midStencilVal, 0xffffffff); // again with the ~0 :P
	// When done, set the states back To something more typical.
	glEnable(GL_DEPTH_TEST);
	glCullFace(GL_BACK);
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
#ifndef GLES2
	glEnable(GL_LIGHTING);
#endif
	glDepthMask(GL_TRUE);
	glCullFace(GL_BACK);
	glStencilFunc(GL_NOTEQUAL, ShadowObject::midStencilVal, 0xffffffff); // !!! was 0xffffff
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

// NOTE: is it the projektion matrix ?
#ifndef GLES2
  // !!! WHY ???
	glPushMatrix();
	  glLoadIdentity();

	  glMatrixMode(GL_MODELVIEW);
	  glPushMatrix();
	    glLoadIdentity();
	    glOrtho(0,1 ,1,0, 0,1);
	    float mat_ambient[]  ={ShadowRed, ShadowGreen, ShadowBlue,.5};
	    float mat_diffuse[]  ={0.0f, 0.0f, 0.0f, 0.5};
	    float mat_specular[] ={0.0f, 0.0f, 0.0f, 0.5};
	    float mat_shininess[]={0.0f}; // upto 128

	    glMaterialfv(GL_FRONT, GL_AMBIENT  , mat_ambient);
	    glMaterialfv(GL_FRONT, GL_DIFFUSE  , mat_diffuse);
	    glMaterialfv(GL_FRONT, GL_SPECULAR , mat_specular);
	    glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);

	    glEnable(GL_BLEND);
	    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	    glDisable(GL_DEPTH_TEST);
	    glColor4f(0.0, 0.0, 0.0, 1.0);

	    if(Global::fx1 != true){
        Global::fx1 = true;
        glDisableClientState(GL_NORMAL_ARRAY);
	    }
	    if(Global::fx2 != false){
        Global::fx2 = false;
        glDisableClientState(GL_COLOR_ARRAY);
	    }

	    GLfloat q3[] = {0,0, 0,1, 1,1, 1,0};
	    glVertexPointer(2, GL_FLOAT, 0, q3);
	    glDrawArrays(GL_TRIANGLE_FAN, 0, 4);

	    glEnable(GL_DEPTH_TEST);
	  glPopMatrix();
	  // NOTE: is it the projektion matrix ?
	  glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

#else
	Global::shader = &Global::shader_stencil;
	glUseProgram(Global::shader->ambient_program);

	glBindBuffer(GL_ARRAY_BUFFER, Global::stencil_vbo);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_DEPTH_TEST);

	glVertexAttribPointer(Global::shader->vposition, 2, GL_FLOAT, GL_FALSE, 0, 0);
	glUniform4f(Global::shader->color,ShadowRed,ShadowGreen,ShadowBlue,.5);
	glEnableVertexAttribArray(Global::shader->vposition);

	glDrawArrays(GL_TRIANGLE_FAN,0,4);
#endif
	glCullFace(GL_BACK);
	glDisable(GL_STENCIL_TEST);
}
